---
title: "Test Implementation Plan: Enhance `khive new-doc` Error Messages"
by: khive-implementer
created: 2025-05-22
updated: 2025-05-22
version: 1.0
doc_type: TI
output_subdir: ti
description: "Test implementation plan for enhancing error messages in the `khive new-doc` CLI command (Issue #103)."
date: 2025-05-22
issue_ref: 103
---

# Guidance

**Purpose** Document the planned and actual test implementation. Clarify unit,
integration, performance, mocking details, and test data.

**When to Use**

- Before/during writing tests, especially if it’s a large feature or
  microservice.
- As a blueprint to ensure coverage is complete.

**Best Practices**

- Keep tests short and focused.
- Use mocking for external calls.
- Outline coverage goals.

---

# Test Implementation Plan: Enhance `khive new-doc` Error Messages (Issue #103)

## 1. Overview

### 1.1 Component Under Test

The `khive new-doc` CLI command, specifically its error handling and messaging
capabilities. The relevant source files are primarily
[`src/khive/cli/khive_new_doc.py`](src/khive/cli/khive_new_doc.py:0) and the
underlying command logic in
[`src/khive/commands/new_doc.py`](src/khive/commands/new_doc.py:0). Tests will
be located in
[`tests/cli/test_khive_new_doc.py`](tests/cli/test_khive_new_doc.py:0).

### 1.2 Test Approach

The primary focus will be on **unit tests** using `pytest`. These tests will
simulate various error conditions and verify that:

1. The command exits with a non-zero status code.
2. The correct, user-friendly error message is printed to stderr for standard
   output.
3. The correct JSON structure (`{"status": "error", "message": "..."}`) is
   printed to stdout when `--json-output` is used.

### 1.3 Key Testing Goals

- Verify clear and actionable error messages for file-already-exists scenarios
  (with and without `--force`).
- Verify clear error messages for template-not-found scenarios.
- Verify error messages for invalid/unwritable destination paths.
- Verify error messages for issues creating destination paths.
- (If applicable) Verify error messages for invalid identifiers.
- Ensure JSON output correctly reflects all error states.
- Maintain existing test coverage for successful command execution.

## 2. Test Environment

### 2.1 Test Framework

```python
# Python
pytest
pytest-mock  # For mocking filesystem and other dependencies
click.testing.CliRunner # For invoking the CLI command in tests
```

### 2.2 Mock Framework

```python
# For Python
unittest.mock (via pytest-mock)
```

We will use `mocker` fixture from `pytest-mock` to patch objects like
`pathlib.Path`, `os.access`, `os.makedirs`, etc.

### 2.3 Test Database

Not applicable for this component, as `khive new-doc` primarily deals with file
system operations and template rendering.

## 3. Unit Tests

All tests will be added to or updated in
[`tests/cli/test_khive_new_doc.py`](tests/cli/test_khive_new_doc.py:0).

### 3.1 Test Suite: `TestNewDocErrorHandling`

#### 3.1.1 Test Case: Output File Already Exists

**Purpose:** Verify correct error message when the target output file already
exists and `--force` is not used. **Setup:**

- Use `CliRunner` to invoke `khive new-doc`.
- Create a temporary directory using `tmp_path` fixture.
- Pre-create a file at the expected output path within `tmp_path`.
- Provide a valid template type and identifier. **Test Implementation (Standard
  Output):**

```python
from click.testing import CliRunner
from khive.cli.khive_cli import cli # Assuming 'cli' is the main click group
import pytest # For tmp_path

def test_new_doc_file_exists_error_no_force(tmp_path, mocker):
    runner = CliRunner()
    output_dir = tmp_path / ".khive" / "reports" / "ip"
    output_dir.mkdir(parents=True, exist_ok=True)
    existing_file = output_dir / "IP-testexists.md"
    existing_file.write_text("Original content")

    # Mock template finding to always succeed for this test
    mocker.patch("khive.commands.new_doc.find_template", return_value=("template_content", "dummy_template.md"))
    mocker.patch("khive.commands.new_doc.render_template", return_value="rendered_content")

    result = runner.invoke(cli, ["new-doc", "IP", "testexists", "--dest", str(tmp_path)])

    assert result.exit_code != 0
    assert "already exists. Use --force to overwrite." in result.stderr
    assert existing_file.read_text() == "Original content" # File not overwritten
```

**Test Implementation (JSON Output):**

```python
def test_new_doc_file_exists_error_no_force_json(tmp_path, mocker):
    runner = CliRunner()
    output_dir = tmp_path / ".khive" / "reports" / "ip"
    output_dir.mkdir(parents=True, exist_ok=True)
    existing_file = output_dir / "IP-testexists.md"
    existing_file.write_text("Original content")

    mocker.patch("khive.commands.new_doc.find_template", return_value=("template_content", "dummy_template.md"))
    mocker.patch("khive.commands.new_doc.render_template", return_value="rendered_content")

    result = runner.invoke(cli, ["new-doc", "IP", "testexists", "--dest", str(tmp_path), "--json-output"])

    assert result.exit_code != 0
    json_output = json.loads(result.stdout)
    assert json_output["status"] == "error"
    assert "already exists. Use --force to overwrite." in json_output["message"]
    assert "IP-testexists.md" in json_output["message"]
    assert existing_file.read_text() == "Original content"
```

**Test Case: Output File Exists with `--force`** **Purpose:** Verify successful
overwrite when `--force` is used. **Setup:** Similar to above, but include
`--force`. **Test Implementation:**

```python
def test_new_doc_file_exists_with_force(tmp_path, mocker):
    runner = CliRunner()
    output_dir = tmp_path / ".khive" / "reports" / "ip"
    output_dir.mkdir(parents=True, exist_ok=True)
    existing_file = output_dir / "IP-testforce.md"
    existing_file.write_text("Original content")

    mocker.patch("khive.commands.new_doc.find_template", return_value=("template_content", "dummy_template.md"))
    mocker.patch("khive.commands.new_doc.render_template", return_value="New rendered content")

    result = runner.invoke(cli, ["new-doc", "IP", "testforce", "--dest", str(tmp_path), "--force"])

    assert result.exit_code == 0
    assert "Document created" in result.stdout # Or similar success message
    assert existing_file.read_text() == "New rendered content" # File overwritten
```

#### 3.1.2 Test Case: Template Not Found

**Purpose:** Verify correct error message when the specified template type or
name cannot be found. **Setup:**

- Use `CliRunner`.
- Mock `khive.commands.new_doc.find_template` to raise `FileNotFoundError` or
  return `None`. **Test Implementation (Standard Output):**

```python
def test_new_doc_template_not_found_error(tmp_path, mocker):
    runner = CliRunner()
    mocker.patch("khive.commands.new_doc.find_template", side_effect=FileNotFoundError("Template 'NonExistentType' not found."))

    result = runner.invoke(cli, ["new-doc", "NonExistentType", "test-id", "--dest", str(tmp_path)])

    assert result.exit_code != 0
    assert "Template 'NonExistentType' not found." in result.stderr # Or a more user-friendly version
```

**Test Implementation (JSON Output):**

```python
def test_new_doc_template_not_found_error_json(tmp_path, mocker):
    runner = CliRunner()
    mocker.patch("khive.commands.new_doc.find_template", side_effect=FileNotFoundError("Template 'NonExistentType' not found."))

    result = runner.invoke(cli, ["new-doc", "NonExistentType", "test-id", "--dest", str(tmp_path), "--json-output"])

    assert result.exit_code != 0
    json_output = json.loads(result.stdout)
    assert json_output["status"] == "error"
    assert "Template 'NonExistentType' not found." in json_output["message"]
```

#### 3.1.3 Test Case: Invalid Identifier (Placeholder - if specific validation is added)

**Purpose:** Verify error for invalid characters or format in the identifier.
**Setup:** `CliRunner`, provide an identifier deemed invalid. **Test
Implementation:** Similar structure to above, checking for specific error
message.

#### 3.1.4 Test Case: Destination Path Not Writable

**Purpose:** Verify error when the destination directory (from `--dest` or
default) is not writable. **Setup:**

- `CliRunner`.
- `tmp_path`.
- Mock `os.access` for the target directory to return `False` for `os.W_OK`.
- Or, mock `Path.mkdir` within the command to raise `PermissionError` when it
  tries to create `output_file.parent`. **Test Implementation (Standard
  Output):**

```python
def test_new_doc_dest_not_writable_error(tmp_path, mocker):
    runner = CliRunner()
    # Simulate a scenario where the parent directory of the output file cannot be written to
    # This might involve mocking Path.mkdir if it's called to create parent dirs
    mocker.patch("pathlib.Path.mkdir", side_effect=PermissionError("Permission denied for path"))
    mocker.patch("khive.commands.new_doc.find_template", return_value=("template_content", "dummy_template.md"))
    mocker.patch("khive.commands.new_doc.render_template", return_value="rendered_content")


    # Attempt to write to a subdirectory of tmp_path that we'll simulate as non-writable
    # by having mkdir fail.
    non_writable_dest = tmp_path / "restricted_dir"

    result = runner.invoke(cli, ["new-doc", "IP", "testperm", "--dest", str(non_writable_dest)])

    assert result.exit_code != 0
    assert "Error creating directory" in result.stderr or "is not writable" in result.stderr # Adjust based on actual message
```

**Test Implementation (JSON Output):**

```python
def test_new_doc_dest_not_writable_error_json(tmp_path, mocker):
    runner = CliRunner()
    mocker.patch("pathlib.Path.mkdir", side_effect=PermissionError("Permission denied for path"))
    mocker.patch("khive.commands.new_doc.find_template", return_value=("template_content", "dummy_template.md"))
    mocker.patch("khive.commands.new_doc.render_template", return_value="rendered_content")

    non_writable_dest = tmp_path / "restricted_dir"
    result = runner.invoke(cli, ["new-doc", "IP", "testperm", "--dest", str(non_writable_dest), "--json-output"])

    assert result.exit_code != 0
    json_output = json.loads(result.stdout)
    assert json_output["status"] == "error"
    assert "Error creating directory" in json_output["message"] or "is not writable" in json_output["message"]
```

#### 3.1.5 Test Case: Cannot Create Destination Path

**Purpose:** Verify error when a non-existent `--dest` path cannot be created
(e.g., intermediate component is a file, or deeper permission issue). **Setup:**

- `CliRunner`.
- `tmp_path`.
- Create a file where a directory is expected in the `--dest` path.
- Mock `pathlib.Path.mkdir` to raise an `OSError` (e.g., `FileExistsError` if an
  intermediate path is a file). **Test Implementation (Standard Output):**

```python
def test_new_doc_cannot_create_dest_path_error(tmp_path, mocker):
    runner = CliRunner()
    # Create a file where a directory is expected
    intermediate_path_as_file = tmp_path / "intermediate_file"
    intermediate_path_as_file.write_text("I am a file")

    dest_path = intermediate_path_as_file / "reports" / "ip"

    mocker.patch("khive.commands.new_doc.find_template", return_value=("template_content", "dummy_template.md"))
    mocker.patch("khive.commands.new_doc.render_template", return_value="rendered_content")
    # Path.mkdir will likely raise FileExistsError or NotADirectoryError if intermediate_path_as_file is a file
    # The actual error handling in new_doc.py should catch this.

    result = runner.invoke(cli, ["new-doc", "IP", "testcreate", "--dest", str(dest_path.parent.parent.parent)]) # Pass base of dest

    assert result.exit_code != 0
    assert "Error creating directory" in result.stderr or "Cannot create destination path" in result.stderr
```

**Test Implementation (JSON Output):**

```python
def test_new_doc_cannot_create_dest_path_error_json(tmp_path, mocker):
    runner = CliRunner()
    intermediate_path_as_file = tmp_path / "intermediate_file"
    intermediate_path_as_file.write_text("I am a file")
    dest_path = intermediate_path_as_file / "reports" / "ip"

    mocker.patch("khive.commands.new_doc.find_template", return_value=("template_content", "dummy_template.md"))
    mocker.patch("khive.commands.new_doc.render_template", return_value="rendered_content")

    result = runner.invoke(cli, ["new-doc", "IP", "testcreate", "--dest", str(dest_path.parent.parent.parent), "--json-output"])

    assert result.exit_code != 0
    json_output = json.loads(result.stdout)
    assert json_output["status"] == "error"
    assert "Error creating directory" in json_output["message"] or "Cannot create destination path" in json_output["message"]
```

## 4. Integration Tests

Not extensively required for this specific enhancement, as the changes are
primarily within the command's internal logic and error reporting, not its
interaction with other distinct services. Existing integration tests (if any)
for `khive new-doc` successful paths should continue to pass.

## 5. API Tests

Not applicable. This is a CLI command.

## 6. Error Handling Tests

This entire TI document focuses on error handling tests for `khive new-doc`.

## 7. Performance Tests

Not applicable for this enhancement. Error message generation should have
negligible performance impact.

## 8. Mock Implementation Details

- `mocker.patch('pathlib.Path.exists', return_value=True)`: To simulate an
  existing output file.
- `mocker.patch('pathlib.Path.is_file', return_value=True)`: To simulate an
  existing output file.
- `mocker.patch('khive.commands.new_doc.find_template', side_effect=FileNotFoundError)`:
  To simulate a template not being found.
- `mocker.patch('os.access', return_value=False)`: To simulate a non-writable
  directory.
- `mocker.patch('pathlib.Path.mkdir', side_effect=PermissionError)`: To simulate
  inability to create a directory due to permissions.
- `mocker.patch('pathlib.Path.mkdir', side_effect=OSError)`: To simulate other
  OS errors during directory creation.

## 9. Test Data

- Valid template types: "IP", "TI", "TDS", etc. (as defined in the project).
- Invalid template type: "NonExistentType".
- Valid identifiers: "test-doc", "001-feature".
- (If applicable) Invalid identifiers based on rules to be defined/checked.
- File content for pre-existing files: "Original content".
- File content for rendered templates: "New rendered content".

## 10. Helper Functions

Standard `pytest` fixtures like `tmp_path` and `mocker` will be used. No
extensive custom test helper functions are anticipated for these specific error
tests beyond what `CliRunner` provides.

## 11. Test Coverage Targets

- **Line Coverage Target:** Aim for >90% for the modified sections in
  [`src/khive/cli/khive_new_doc.py`](src/khive/cli/khive_new_doc.py:0) and
  [`src/khive/commands/new_doc.py`](src/khive/commands/new_doc.py:0) related to
  error handling.
- **Branch Coverage Target:** Ensure all new error handling branches are covered
  by tests.
- Overall project coverage targets as per project standards should be
  maintained.

## 12. Continuous Integration

Existing CI setup should run these new tests. Ensure `pytest --cov` reports
coverage accurately.

## 13. Notes and Caveats

### 13.1 Known Limitations

- Testing exact permission errors across different OS platforms can be tricky;
  mocks will simulate the OS behavior.
- The exact wording of error messages might be refined during implementation.
  Tests will need to be updated accordingly.

### 13.2 Future Improvements

- Consider more granular error codes in JSON output if needed by consuming
  tools.
