---
title: "Test Implementation Plan: Reader Microservice - Document Processing Pipeline"
by: "khive-implementer"
created: "2025-05-22"
updated: "2025-05-22"
version: "1.0"
doc_type: "TI"
issue_ref: "27"
output_subdir: "ti"
description: "Test implementation plan for the background worker and document processing pipeline for the Khive Reader Microservice (Issue #27)."
---

# Guidance

**Purpose** Document the planned and actual test implementation. Clarify unit,
integration, performance, mocking details, and test data.

**When to Use**

- Before/during writing tests, especially if it’s a large feature or
  microservice.
- As a blueprint to ensure coverage is complete.

**Best Practices**

- Keep tests short and focused.
- Use mocking for external calls.
- Outline coverage goals.

---

# Test Implementation Plan: Reader Microservice - Document Processing Pipeline

## 1. Overview

### 1.1 Component Under Test

This test plan covers the new components introduced for the Reader
Microservice's document processing pipeline, as specified in Issue #27. This
includes:

- `src/khive/reader/tasks/queue.py` (`AsyncTaskQueue`)
- `src/khive/reader/processing/text_extraction.py` (`TextExtractor`)
- `src/khive/reader/processing/chunking.py` (`DocumentChunker`)
- `src/khive/reader/processing/embeddings.py` (`EmbeddingGenerator`)
- `src/khive/reader/services/processing_service.py`
  (`DocumentProcessingService`)
- Task registration in `src/khive/reader/tasks/__init__.py`
- Modifications to the existing Ingestion Service.

### 1.2 Test Approach

A combination of unit tests and integration tests will be employed.

- **Unit Tests:** Each new module and class will be tested in isolation,
  focusing on its specific logic and error handling. Mocking will be used
  extensively for external dependencies (APIs, repositories, other services).
- **Integration Tests:** Will verify the interaction between the newly developed
  components, such as the `DocumentProcessingService` orchestrating the
  `TextExtractor`, `DocumentChunker`, and `EmbeddingGenerator`. Further
  integration tests will cover the flow from the updated Ingestion Service
  through the `AsyncTaskQueue` to the `DocumentProcessingService`.

### 1.3 Key Testing Goals

- Verify correct functionality of each processing step: text extraction for all
  supported formats, text chunking, and embedding generation (OpenAI and
  fallback).
- Ensure the `DocumentProcessingService` correctly orchestrates the pipeline and
  updates document statuses.
- Validate robust error handling at each stage of the pipeline.
- Confirm the `AsyncTaskQueue` correctly manages and dispatches tasks.
- Verify the updated Ingestion Service successfully queues documents for
  processing.
- Achieve >=80% unit test coverage for all new code.

## 2. Test Environment

### 2.1 Test Framework

```python
# Python
pytest
pytest-asyncio  # For testing asynchronous code
pytest-mock     # For mocker fixture
pytest-cov      # For coverage reporting
```

### 2.2 Mock Framework

```python
# For Python
unittest.mock   # Standard library
pytest-mock     # Provides mocker fixture, often preferred with pytest
```

### 2.3 Test Database

Repositories (`DocumentRepository`, `DocumentChunkRepository`) will be mocked
for unit tests. For integration tests focusing on the pipeline logic, these will
also be mocked to avoid actual database interactions and keep tests fast and
isolated. If full E2E tests involving a database were in scope, an ephemeral
test database (e.g., SQLite in-memory or a test container) would be considered,
but that's beyond this TI's primary focus.

## 3. Unit Tests

The unit tests will be structured according to the test groups defined in the
Implementation Plan (IP-27.md), Section 3.1. Below are illustrative examples for
some key test cases.

### 3.1 Test Suite: `src/khive/reader/tasks/queue.py` (`AsyncTaskQueue`)

#### 3.1.1 Test Case: `test_submit_and_get_task`

**Purpose:** Verify basic task submission and retrieval. **Setup:**

```python
import pytest
from src.khive.reader.tasks.queue import AsyncTaskQueue # Assuming path

@pytest.fixture
async def task_queue():
    return AsyncTaskQueue()
```

**Test Implementation:**

```python
async def test_submit_and_get_task(task_queue: AsyncTaskQueue):
    # Arrange
    test_item = "document_id_123"
    await task_queue.submit_task(test_item)

    # Act
    retrieved_item = await task_queue.get_task()

    # Assert
    assert retrieved_item == test_item
    assert task_queue.qsize() == 0 # Assuming get_task also implies task is taken for processing
```

#### 3.1.2 Test Case: `test_get_task_waits_for_item`

(Details in IP-27.md, Section 3.1.1, UT-Q4)

### 3.2 Test Suite: `src/khive/reader/processing/text_extraction.py` (`TextExtractor`)

#### 3.2.1 Test Case: `test_extract_text_pdf_valid`

**Purpose:** Verify PDF text extraction. **Setup:** Create a fixture for a
sample PDF file.

```python
from pathlib import Path
import pytest
from src.khive.reader.processing.text_extraction import TextExtractor # Assuming path

@pytest.fixture
def sample_pdf_path(tmp_path: Path) -> Path:
    pdf_content = b"%PDF-1.4\n1 0 obj<</Type/Catalog/Pages 2 0 R>>endobj\n2 0 obj<</Type/Pages/Count 1/Kids[3 0 R]>>endobj\n3 0 obj<</Type/Page/MediaBox[0 0 612 792]/Parent 2 0 R/Resources<<>>/Contents 4 0 R>>endobj\n4 0 obj<</Length 38>>stream\nBT /F1 24 Tf 100 700 Td (Hello PDF) Tj ET\nendstream\nendobj\nxref\n0 5\n0000000000 65535 f\n0000000010 00000 n\n0000000059 00000 n\n0000000118 00000 n\n0000000212 00000 n\ntrailer<</Size 5/Root 1 0 R>>\nstartxref\n276\n%%EOF" # Simplified PDF content
    file_path = tmp_path / "sample.pdf"
    file_path.write_bytes(pdf_content)
    return file_path

@pytest.fixture
def text_extractor():
    return TextExtractor()
```

**Test Implementation:**

```python
def test_extract_text_pdf_valid(text_extractor: TextExtractor, sample_pdf_path: Path):
    # Act
    # Note: PyPDF2 might require a more complex valid PDF for actual text.
    # This test might need a real small PDF or more robust mock content.
    # For now, assuming a simplified scenario or that PyPDF2 handles this.
    # If PyPDF2 struggles with minimal PDFs, we'd use a known-good small PDF file.
    try:
        text = text_extractor.extract_text(sample_pdf_path, "application/pdf")
        # Assert
        assert "Hello PDF" in text # This depends on PyPDF2's capability with the minimal PDF
    except Exception as e:
        # Depending on PyPDF2's behavior with such a minimal PDF,
        # it might raise an error or return empty. Adjust assertion accordingly.
        # For a real test, use a known-good small PDF.
        pytest.skip(f"Skipping due to PyPDF2 behavior with minimal PDF: {e}")
```

_(Similar test cases for DOCX, HTML, TXT, and error conditions as per IP-27.md,
Section 3.1.2)_

### 3.3 Test Suite: `src/khive/reader/processing/chunking.py` (`DocumentChunker`)

(Test cases as per IP-27.md, Section 3.1.3)

### 3.4 Test Suite: `src/khive/reader/processing/embeddings.py` (`EmbeddingGenerator`)

#### 3.4.1 Test Case: `test_generate_embeddings_openai_success`

**Purpose:** Verify successful embedding generation using mocked OpenAI.
**Setup:**

```python
import pytest
from unittest.mock import AsyncMock # If openai client is async
from src.khive.reader.processing.embeddings import EmbeddingGenerator # Assuming path

@pytest.fixture
def embedding_generator(mocker):
    # Mock the OpenAI client if it's instantiated within EmbeddingGenerator
    # or pass a mocked client if injected.
    # For this example, assume it uses openai.Embedding directly.
    mock_openai_create = mocker.patch("openai.resources.Embeddings.create", new_callable=AsyncMock) # Adjust path if needed
    mock_openai_create.return_value.data = [mocker.Mock(embedding=[0.1, 0.2, 0.3])]
    return EmbeddingGenerator(openai_api_key="fake_key") # Or however API key is passed
```

**Test Implementation:**

```python
async def test_generate_embeddings_openai_success(embedding_generator: EmbeddingGenerator, mocker):
    # Arrange
    texts = ["hello world"]

    # Act
    embeddings = await embedding_generator.generate_embeddings(texts)

    # Assert
    assert embeddings == [[0.1, 0.2, 0.3]]
    embedding_generator.openai_client.embeddings.create.assert_called_once_with(
        model=embedding_generator.openai_model, # or the default model
        input=texts
    )
```

_(Test cases for fallback and errors as per IP-27.md, Section 3.1.4)_

### 3.5 Test Suite: `src/khive/reader/services/processing_service.py` (`DocumentProcessingService`)

#### 3.5.1 Test Case: `test_process_document_happy_path`

**Purpose:** Verify the successful end-to-end processing flow within the
service. **Setup:** Mock all dependencies (`ObjectStorageClient`,
`TextExtractor`, `DocumentChunker`, `EmbeddingGenerator`, `DocumentRepository`,
`DocumentChunkRepository`). **Test Implementation:** (Conceptual)

```python
# Conceptual structure
async def test_process_document_happy_path(processing_service, mock_doc_repo, mock_chunk_repo, /* other mocks */):
    # Arrange: Configure mocks to return successful results for each step
    # mock_object_storage.download_file_to_temp.return_value = Path("fake_doc.pdf")
    # mock_text_extractor.extract_text.return_value = "extracted text"
    # mock_document_chunker.chunk_text.return_value = ["chunk1", "chunk2"]
    # mock_embedding_generator.generate_embeddings.return_value = [[0.1], [0.2]]
    # mock_doc_repo.get_by_id.return_value = MockDocument(id="doc1", status="PENDING", file_path_in_object_store="remote/doc1.pdf")

    document_id = "doc1"

    # Act
    await processing_service.process_document(document_id)

    # Assert:
    # - All mock methods were called in the correct order with correct arguments.
    # - mock_doc_repo.update_status was called multiple times with correct statuses (DOWNLOADING, EXTRACTING, CHUNKING, EMBEDDING, PROCESSED).
    # - mock_chunk_repo.create_many_chunks_with_embeddings was called with correct chunk and embedding data.
    # - mock_doc_repo.set_processed_text_path was called (if applicable).
```

_(Test cases for error handling at each step as per IP-27.md, Section 3.1.5)_

## 4. Integration Tests

### 4.1 Test Suite: Full Document Processing Pipeline (Task Queue to Processing Service)

#### 4.1.1 Test Case: `test_pdf_processing_via_queue`

**Purpose:** Verify a PDF document ID submitted to the queue is processed by the
service. **Components Involved:** `AsyncTaskQueue`, `process_document_task`
(worker function), `DocumentProcessingService`, and its direct dependencies
(mocked: `ObjectStorageClient`, `TextExtractor`, `DocumentChunker`,
`EmbeddingGenerator`, Repositories). **Setup:**

- Instantiate a real `AsyncTaskQueue`.
- Create a `process_document_task` function that uses a
  `DocumentProcessingService` instance (with mocked dependencies).
- Start a "worker" asyncio task that continuously calls `queue.get_task()` and
  then `process_document_task(doc_id)`.
- Provide a sample PDF (or mock its download and extraction). **Test
  Implementation:** (Conceptual)

```python
# Conceptual structure
async def test_pdf_processing_via_queue(event_loop, task_queue, mock_processing_service_factory):
    # Arrange
    document_id = "pdf_doc_id_1"
    # mock_processing_service_factory creates a DocumentProcessingService with all its deps mocked
    # and allows assertions on these mocks later.
    # The process_document_task would use this factory.

    async def worker():
        task_item = await task_queue.get_task()
        if task_item:
            service_instance = mock_processing_service_factory(task_item) # or however service gets doc_id
            await service_instance.process_document(task_item) # or process_document_task(task_item)
            task_queue.task_done()

    worker_task = event_loop.create_task(worker())

    # Act
    await task_queue.submit_task(document_id)
    await task_queue.join() # Wait for the task to be processed
    worker_task.cancel() # Clean up worker

    # Assert:
    # - Assert that the mock_processing_service's process_document method was called with document_id.
    # - Further assertions on the internal mocks of the processing_service (e.g., status updates, chunk storage)
    #   can be done if the factory provides access to them.
```

_(Similar tests for DOCX, and OpenAI failure/fallback as per IP-27.md, Section
3.2.1)_

### 4.2 Test Suite: Ingestion Service Integration

(Test cases as per IP-27.md, Section 3.2.2)

## 5. API Tests

Not directly applicable for these backend processing components, as they don't
expose external APIs themselves. API tests would be relevant for the Ingestion
Service's endpoint if it's being modified.

## 6. Error Handling Tests

Error handling is a key part of the unit tests for `TextExtractor`,
`EmbeddingGenerator`, and `DocumentProcessingService`. Each component will be
tested for its resilience to failures in its dependencies or invalid inputs,
ensuring appropriate exceptions are raised or errors are logged and document
statuses are updated correctly. (Refer to specific error handling test cases in
IP-27.md, e.g., UT-TE5, UT-EG5, UT-PS2-PS5).

## 7. Performance Tests

Performance testing is not in the immediate scope of this TI but could be a
future improvement. If implemented, it would focus on:

- Throughput of the `AsyncTaskQueue`.
- Processing time per document for `DocumentProcessingService` under various
  loads and document sizes.
- Benchmarking `TextExtractor` and `EmbeddingGenerator` for different file
  types/sizes.

## 8. Mock Implementation Details

Mocks will primarily use `pytest-mock`'s `mocker` fixture or `unittest.mock`.
Example for `DocumentRepository`:

```python
from unittest.mock import MagicMock

@pytest.fixture
def mock_document_repository(mocker):
    mock_repo = mocker.MagicMock(spec=DocumentRepository) # Assuming DocumentRepository is defined
    mock_repo.get_by_id = mocker.AsyncMock(return_value=MagicMock(id="doc1", status="PENDING", file_path_in_object_store="path/to/file.pdf"))
    mock_repo.update_status = mocker.AsyncMock()
    mock_repo.set_processed_text_path = mocker.AsyncMock()
    return mock_repo
```

Similar mock fixtures will be created for `ObjectStorageClient`,
`DocumentChunkRepository`, `TextExtractor`, `DocumentChunker`,
`EmbeddingGenerator`, `openai.Embedding`, and `SentenceTransformer` as needed
for different test suites.

## 9. Test Data

- Sample files: Small, valid PDF, DOCX, HTML, TXT files.
- Corrupted/empty files for error testing.
- Sample text snippets of varying lengths for chunking tests.
- Expected extracted text for validation.
- Pre-computed (or expected format of) embeddings for assertion.

Test data files will be stored likely in a `tests/fixtures/files` directory.

## 10. Helper Functions

- Fixtures for creating mock objects (as shown in section 8).
- Fixtures for providing paths to sample test files.
- Potentially, helper functions to create `Document` model instances for
  repository mocks.

## 11. Test Coverage Targets

- **Line Coverage Target:** >=80% for all new modules:
  - `src/khive/reader/tasks/queue.py`
  - `src/khive/reader/processing/text_extraction.py`
  - `src/khive/reader/processing/chunking.py`
  - `src/khive/reader/processing/embeddings.py`
  - `src/khive/reader/services/processing_service.py`
- **Branch Coverage Target:** Aim for >=75%, focusing on conditional logic in
  error handling and fallbacks.
- Coverage reports will be generated using `pytest-cov`.

## 12. Continuous Integration

CI pipeline (e.g., GitHub Actions) should be configured to:

1. Install dependencies (including test dependencies).
2. Run `uv run pre-commit run --all-files`.
3. Run all tests using
   `uv run pytest tests/ --cov=src/khive/reader --cov-report=xml`.
4. Upload coverage reports (e.g., to Codecov).

(YAML example provided in template is a good starting point).

## 13. Notes and Caveats

### 13.1 Known Limitations

- Initial tests for `TextExtractor` with minimal PDF/DOCX content might require
  actual small, valid files rather than byte-string mocks if libraries are
  sensitive.
- Full testing of OpenAI API calls relies on robust mocking; actual API
  interaction testing is out of scope for unit/integration tests here.

### 13.2 Future Improvements

- Add performance benchmark tests.
- Expand integration tests to include a real (test) database if deemed
  necessary.
- Test with a wider variety of complex and malformed documents.
