---
title: Test Implementation for Robust Async API Client
by: khive-implementer
created: 2025-05-18
updated: 2025-05-18
version: 1.0
doc_type: TI
output_subdir: ti
description: Comprehensive test implementation for the robust async API client with resource management, rate limiting, and error handling
date: 2025-05-18
---

# Test Implementation Plan: Robust Async API Client

## 1. Overview

### 1.1 Component Under Test

The Robust Async API Client implementation, which includes:

- AsyncAPIClient for HTTP interactions with proper resource management
- TokenBucketRateLimiter for rate limiting
- AsyncExecutor and RateLimitedExecutor for concurrency control
- CircuitBreaker and retry mechanisms for resilience

This test implementation covers all components defined in the Implementation
Plan (IP-81.md) and ensures they work correctly both individually and together.

### 1.2 Test Approach

The test approach follows Test-Driven Development (TDD) principles:

- **Unit Tests**: Testing individual components in isolation with mocked
  dependencies
- **Integration Tests**: Testing components working together
- **Error Handling Tests**: Specific tests for error scenarios
- **Resource Management Tests**: Tests to ensure proper resource cleanup

### 1.3 Key Testing Goals

- Verify proper resource management (async context managers)
- Ensure correct error handling and propagation
- Validate rate limiting behavior
- Confirm concurrency control works as expected
- Test resilience patterns (circuit breaker, retry)
- Achieve >80% test coverage

## 2. Test Environment

### 2.1 Test Framework

```
pytest
pytest-asyncio
pytest-mock
pytest-cov
```

### 2.2 Mock Framework

```
unittest.mock
pytest-mock
```

### 2.3 Test Database

Not applicable for this component.

## 3. Unit Tests

### 3.1 Test Suite: AsyncAPIClient

#### 3.1.1 Test Case: Initialization

**Purpose**: Verify that AsyncAPIClient initializes correctly with the provided
parameters.

**Setup**: None required.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_async_api_client_init():
    # Arrange
    base_url = "https://api.example.com"
    timeout = 10.0
    headers = {"User-Agent": "Test"}

    # Act
    client = AsyncAPIClient(
        base_url=base_url,
        timeout=timeout,
        headers=headers
    )

    # Assert
    assert client.base_url == base_url
    assert client.timeout == timeout
    assert client.headers == headers
    assert client._client is None
    assert client._closed is False
```

#### 3.1.2 Test Case: Context Manager

**Purpose**: Verify that AsyncAPIClient properly implements the async context
manager protocol.

**Setup**: Mock httpx.AsyncClient.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_async_api_client_context_manager():
    # Arrange
    base_url = "https://api.example.com"
    mock_session = AsyncMock(spec=httpx.AsyncClient)

    # Act & Assert
    with patch('httpx.AsyncClient', return_value=mock_session):
        async with AsyncAPIClient(base_url=base_url) as client:
            # Assert client was initialized correctly
            assert client.base_url == base_url
            assert client._client is not None

        # Assert session was closed
        mock_session.aclose.assert_called_once()
```

#### 3.1.3 Test Case: HTTP Methods

**Purpose**: Verify that HTTP methods (get, post, etc.) work correctly.

**Setup**: Mock httpx.AsyncClient and response.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_async_api_client_get():
    # Arrange
    base_url = "https://api.example.com"
    mock_session = AsyncMock(spec=httpx.AsyncClient)
    mock_response = MagicMock()
    mock_response.json.return_value = {"data": "test"}
    mock_session.request.return_value = mock_response

    # Act
    with patch('httpx.AsyncClient', return_value=mock_session):
        async with AsyncAPIClient(base_url=base_url) as client:
            result = await client.get("/test", params={"key": "value"})

    # Assert
    mock_session.request.assert_called_once_with(
        "GET",
        "/test",
        params={"key": "value"}
    )
    assert result == {"data": "test"}
```

#### 3.1.4 Test Case: Error Handling

**Purpose**: Verify that different types of errors are handled correctly.

**Setup**: Mock httpx.AsyncClient to raise different exceptions.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_async_api_client_connection_error():
    # Arrange
    base_url = "https://api.example.com"
    mock_session = AsyncMock(spec=httpx.AsyncClient)
    mock_session.request.side_effect = httpx.ConnectError("Connection failed")

    # Act & Assert
    with patch('httpx.AsyncClient', return_value=mock_session):
        async with AsyncAPIClient(base_url=base_url) as client:
            with pytest.raises(ConnectionError) as excinfo:
                await client.get("/test")

    # Assert
    assert "Connection error: Connection failed" in str(excinfo.value)
```

### 3.2 Test Suite: TokenBucketRateLimiter

#### 3.2.1 Test Case: Initialization

**Purpose**: Verify that TokenBucketRateLimiter initializes correctly.

**Setup**: None required.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_token_bucket_rate_limiter_init():
    # Arrange
    rate = 10
    period = 1.0
    max_tokens = 15

    # Act
    limiter = TokenBucketRateLimiter(
        rate=rate,
        period=period,
        max_tokens=max_tokens
    )

    # Assert
    assert limiter.rate == rate
    assert limiter.period == period
    assert limiter.max_tokens == max_tokens
    assert limiter.tokens == max_tokens
```

#### 3.2.2 Test Case: Token Acquisition

**Purpose**: Verify that tokens are acquired correctly and wait times are
calculated properly.

**Setup**: Mock time.monotonic to control time progression.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_token_bucket_rate_limiter_acquire_tokens_available():
    # Arrange
    rate = 10
    period = 1.0
    limiter = TokenBucketRateLimiter(rate=rate, period=period)
    limiter.tokens = 5  # Start with 5 tokens

    # Mock _refill to do nothing
    with patch.object(limiter, '_refill', AsyncMock()):
        # Act
        wait_time = await limiter.acquire(tokens=3)

        # Assert
        assert wait_time == 0.0
        assert limiter.tokens == 2  # 5 - 3 = 2
```

#### 3.2.3 Test Case: Rate Limiting Execution

**Purpose**: Verify that execute method applies rate limiting correctly.

**Setup**: Mock acquire method and asyncio.sleep.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_token_bucket_rate_limiter_execute_with_wait():
    # Arrange
    rate = 10
    period = 1.0
    limiter = TokenBucketRateLimiter(rate=rate, period=period)

    # Mock acquire to return 0.2 (wait 0.2 seconds)
    with patch.object(limiter, 'acquire', AsyncMock(return_value=0.2)):
        # Mock asyncio.sleep
        mock_sleep = AsyncMock()

        # Mock the function to be executed
        mock_func = AsyncMock(return_value="result")

        # Act
        with patch('asyncio.sleep', mock_sleep):
            result = await limiter.execute(mock_func, "arg1", "arg2", kwarg1="value1")

        # Assert
        mock_sleep.assert_called_once_with(0.2)
        mock_func.assert_called_once_with("arg1", "arg2", kwarg1="value1")
        assert result == "result"
```

### 3.3 Test Suite: AsyncExecutor

#### 3.3.1 Test Case: Initialization

**Purpose**: Verify that AsyncExecutor initializes correctly.

**Setup**: None required.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_async_executor_init():
    # Arrange & Act
    executor = AsyncExecutor(max_concurrency=5)

    # Assert
    assert executor.semaphore is not None
    assert executor._active_tasks == {}
```

#### 3.3.2 Test Case: Execution with Concurrency Control

**Purpose**: Verify that execute method applies concurrency control correctly.

**Setup**: Mock function to execute.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_async_executor_execute():
    # Arrange
    executor = AsyncExecutor(max_concurrency=5)
    mock_func = AsyncMock(return_value="result")

    # Act
    result = await executor.execute(mock_func, "arg1", "arg2", kwarg1="value1")

    # Assert
    mock_func.assert_called_once_with("arg1", "arg2", kwarg1="value1")
    assert result == "result"
    assert len(executor._active_tasks) == 0  # Task should be removed after completion
```

### 3.4 Test Suite: CircuitBreaker

#### 3.4.1 Test Case: State Transitions

**Purpose**: Verify that CircuitBreaker transitions between states correctly.

**Setup**: Mock time.time for controlled testing.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_circuit_breaker_open_after_threshold():
    # Arrange
    breaker = CircuitBreaker(failure_threshold=3, recovery_time=30.0)
    mock_func = AsyncMock(side_effect=ValueError("Test error"))

    # Act & Assert
    for i in range(3):
        with pytest.raises(ValueError):
            await breaker.execute(mock_func)

    # Assert
    assert breaker.state == CircuitState.OPEN
    assert breaker.failure_count == 3
```

#### 3.4.2 Test Case: Request Rejection

**Purpose**: Verify that CircuitBreaker rejects requests when open.

**Setup**: Set CircuitBreaker to OPEN state.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_circuit_breaker_rejects_when_open():
    # Arrange
    breaker = CircuitBreaker(failure_threshold=3, recovery_time=30.0)
    breaker.state = CircuitState.OPEN
    breaker.last_failure_time = time.time()  # Set to current time
    mock_func = AsyncMock(return_value="result")

    # Act & Assert
    with pytest.raises(CircuitBreakerOpenError) as excinfo:
        await breaker.execute(mock_func)

    # Assert
    assert "Circuit breaker is open" in str(excinfo.value)
    mock_func.assert_not_called()
```

### 3.5 Test Suite: retry_with_backoff

#### 3.5.1 Test Case: Retry Behavior

**Purpose**: Verify that retry_with_backoff retries failed operations with
exponential backoff.

**Setup**: Mock function that fails a certain number of times before succeeding.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_retry_with_backoff_success_after_retries():
    # Arrange
    mock_func = AsyncMock(side_effect=[
        ValueError("Error 1"),
        ValueError("Error 2"),
        "result"
    ])

    # Mock asyncio.sleep to avoid actual delays
    mock_sleep = AsyncMock()

    # Act
    with patch('asyncio.sleep', mock_sleep):
        result = await retry_with_backoff(
            mock_func, "arg1", "arg2",
            kwarg1="value1",
            max_retries=3,
            base_delay=1.0,
            jitter=False  # Disable jitter for predictable testing
        )

    # Assert
    assert mock_func.call_count == 3
    assert result == "result"
    assert mock_sleep.call_count == 2
    # First retry should wait base_delay (1.0)
    # Second retry should wait base_delay * backoff_factor (1.0 * 2.0 = 2.0)
    mock_sleep.assert_any_call(1.0)
    mock_sleep.assert_any_call(2.0)
```

## 4. Integration Tests

### 4.1 Test Suite: API Client with Rate Limiting

**Purpose**: Verify that AsyncAPIClient works correctly with
TokenBucketRateLimiter.

**Setup**: Mock httpx.AsyncClient and set up a RateLimitedExecutor.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_api_client_with_rate_limiting():
    # Arrange
    base_url = "https://api.example.com"
    mock_session = AsyncMock(spec=httpx.AsyncClient)
    mock_response = MagicMock()
    mock_response.json.return_value = {"data": "test"}
    mock_session.request.return_value = mock_response

    # Create rate limiter and client
    limiter = TokenBucketRateLimiter(rate=10, period=1.0)

    # Act
    with patch('httpx.AsyncClient', return_value=mock_session):
        async with AsyncAPIClient(base_url=base_url) as client:
            # Execute with rate limiting
            result = await limiter.execute(
                client.get,
                "/test",
                params={"key": "value"}
            )

    # Assert
    mock_session.request.assert_called_once()
    assert result == {"data": "test"}
```

### 4.2 Test Suite: API Client with Circuit Breaker

**Purpose**: Verify that AsyncAPIClient works correctly with CircuitBreaker.

**Setup**: Mock httpx.AsyncClient to fail and set up a CircuitBreaker.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_api_client_with_circuit_breaker():
    # Arrange
    base_url = "https://api.example.com"
    mock_session = AsyncMock(spec=httpx.AsyncClient)
    mock_session.request.side_effect = [
        httpx.HTTPStatusError(
            "500 Internal Server Error",
            request=MagicMock(),
            response=MagicMock(status_code=500, json=lambda: {"error": "Server error"})
        ),
        httpx.HTTPStatusError(
            "500 Internal Server Error",
            request=MagicMock(),
            response=MagicMock(status_code=500, json=lambda: {"error": "Server error"})
        ),
        MagicMock(json=lambda: {"data": "success"})
    ]

    # Create circuit breaker
    breaker = CircuitBreaker(failure_threshold=2, recovery_time=0.1)

    # Act & Assert
    with patch('httpx.AsyncClient', return_value=mock_session):
        async with AsyncAPIClient(base_url=base_url) as client:
            # First two calls should fail
            for _ in range(2):
                with pytest.raises(ServerError):
                    await breaker.execute(client.get, "/test")

            # Circuit should be open now
            with pytest.raises(CircuitBreakerOpenError):
                await breaker.execute(client.get, "/test")

            # Wait for recovery time
            await asyncio.sleep(0.2)

            # Circuit should be half-open and allow the next call
            result = await breaker.execute(client.get, "/test")
            assert result == {"data": "success"}

            # Circuit should be closed again
            assert breaker.state == CircuitState.CLOSED
```

### 4.3 Test Suite: Complete Integration

**Purpose**: Verify that all components work together correctly.

**Setup**: Set up all components and mock external dependencies.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_complete_integration():
    # Arrange
    base_url = "https://api.example.com"
    mock_session = AsyncMock(spec=httpx.AsyncClient)
    mock_response = MagicMock()
    mock_response.json.return_value = {"data": "test"}
    mock_session.request.return_value = mock_response

    # Create components
    limiter = TokenBucketRateLimiter(rate=10, period=1.0)
    executor = AsyncExecutor(max_concurrency=5)
    rate_limited_executor = RateLimitedExecutor(rate=10, period=1.0, max_concurrency=5)
    breaker = CircuitBreaker(failure_threshold=3, recovery_time=30.0)

    # Act
    with patch('httpx.AsyncClient', return_value=mock_session):
        async with AsyncAPIClient(base_url=base_url) as client:
            # Execute with all components
            result = await breaker.execute(
                rate_limited_executor.execute,
                client.get,
                "/test",
                params={"key": "value"}
            )

    # Assert
    mock_session.request.assert_called_once()
    assert result == {"data": "test"}
```

## 5. API Tests

Not applicable for this component as it is a client library, not an API server.

## 6. Error Handling Tests

### 6.1 Test Suite: API Client Error Handling

**Purpose**: Verify that AsyncAPIClient handles and propagates errors correctly.

**Setup**: Mock httpx.AsyncClient to raise different types of exceptions.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_api_client_error_handling():
    # Arrange
    base_url = "https://api.example.com"
    mock_session = AsyncMock(spec=httpx.AsyncClient)

    # Test different error types
    error_scenarios = [
        (httpx.ConnectError("Connection failed"), ConnectionError),
        (httpx.TimeoutException("Request timed out"), TimeoutError),
        (httpx.HTTPStatusError(
            "429 Too Many Requests",
            request=MagicMock(),
            response=MagicMock(
                status_code=429,
                json=lambda: {"error": "Rate limit exceeded"},
                headers={"Retry-After": "60"}
            )
        ), RateLimitError),
        (httpx.HTTPStatusError(
            "401 Unauthorized",
            request=MagicMock(),
            response=MagicMock(
                status_code=401,
                json=lambda: {"error": "Unauthorized"}
            )
        ), AuthenticationError),
        (httpx.HTTPStatusError(
            "404 Not Found",
            request=MagicMock(),
            response=MagicMock(
                status_code=404,
                json=lambda: {"error": "Not found"}
            )
        ), ResourceNotFoundError),
        (httpx.HTTPStatusError(
            "500 Internal Server Error",
            request=MagicMock(),
            response=MagicMock(
                status_code=500,
                json=lambda: {"error": "Server error"}
            )
        ), ServerError),
    ]

    # Act & Assert
    for error, expected_exception in error_scenarios:
        mock_session.request.side_effect = error

        with patch('httpx.AsyncClient', return_value=mock_session):
            async with AsyncAPIClient(base_url=base_url) as client:
                with pytest.raises(expected_exception):
                    await client.get("/test")
```

### 6.2 Test Suite: Resource Cleanup on Error

**Purpose**: Verify that resources are properly cleaned up when errors occur.

**Setup**: Mock httpx.AsyncClient and force exceptions.

**Test Implementation**:

```python
@pytest.mark.asyncio
async def test_resource_cleanup_on_error():
    # Arrange
    base_url = "https://api.example.com"
    mock_session = AsyncMock(spec=httpx.AsyncClient)

    # Act & Assert
    with patch('httpx.AsyncClient', return_value=mock_session):
        try:
            async with AsyncAPIClient(base_url=base_url) as client:
                # Force an exception
                raise ValueError("Test exception")
        except ValueError:
            pass

        # Assert session was closed despite the exception
        mock_session.aclose.assert_called_once()
```

## 7. Performance Tests

Not implemented in this phase. Will be added in a future update if needed.

## 8. Mock Implementation Details

The tests use the following mock implementations:

```python
# Mock httpx.AsyncClient
mock_session = AsyncMock(spec=httpx.AsyncClient)
mock_session.request.return_value = mock_response
mock_response = MagicMock()
mock_response.json.return_value = {"data": "test"}

# Mock time.monotonic for rate limiter tests
with patch('time.monotonic', side_effect=[0.0, 0.5]):
    # Test code that uses time.monotonic

# Mock asyncio.sleep to avoid actual delays
mock_sleep = AsyncMock()
with patch('asyncio.sleep', mock_sleep):
    # Test code that uses asyncio.sleep
```

## 9. Test Data

```python
# Test URLs
base_url = "https://api.example.com"
test_endpoints = ["/test", "/users", "/items"]

# Test request data
test_request_data = {
    "key": "value",
    "nested": {
        "key": "value"
    },
    "array": [1, 2, 3]
}

# Test response data
test_response_data = {
    "data": "test",
    "status": "success",
    "timestamp": "2025-05-18T12:00:00Z"
}
```

## 10. Helper Functions

```python
# Helper function to create a mock HTTP response
def create_mock_response(status_code=200, json_data=None, headers=None):
    mock_response = MagicMock()
    mock_response.status_code = status_code
    mock_response.json.return_value = json_data or {}
    mock_response.headers = headers or {}
    mock_response.text = str(json_data)
    return mock_response

# Helper function to create a mock HTTP error
def create_mock_http_error(status_code, error_message, headers=None):
    mock_request = MagicMock()
    mock_response = MagicMock(
        status_code=status_code,
        json=lambda: {"error": error_message},
        text=error_message,
        headers=headers or {}
    )
    return httpx.HTTPStatusError(
        f"{status_code} Error",
        request=mock_request,
        response=mock_response
    )
```

## 11. Test Coverage Targets

- **Line Coverage Target:** 85%
- **Branch Coverage Target:** 80%
- **Critical Components:**
  - AsyncAPIClient: 90% coverage
  - CircuitBreaker: 90% coverage
  - Resource management code: 100% coverage

## 12. Continuous Integration

The tests will be run as part of the project's CI pipeline using the existing
configuration.

## 13. Notes and Caveats

### 13.1 Known Limitations

- Some tests use mocked time functions, which may not perfectly simulate
  real-world timing behavior.
- Integration tests with real HTTP servers are not included in this phase.

### 13.2 Future Improvements

- Add integration tests with real HTTP servers
- Add performance tests for rate limiting and concurrency control
- Add more comprehensive error scenario tests
- Add tests for edge cases like very high concurrency or very low rate limits
