---
title: Test Implementation Plan for Bounded Async Queue with Backpressure
by: khive-implementer
created: 2025-05-18
updated: 2025-05-18
version: 1.0
doc_type: TI
output_subdir: ti
description: Test implementation plan for a bounded async queue with backpressure for API requests
date: 2025-05-18
---

# Test Implementation Plan: Bounded Async Queue with Backpressure

## 1. Overview

### 1.1 Component Under Test

This test plan covers the implementation of a bounded async queue with
backpressure for API requests in the khive resource control framework. The main
components under test are:

- `BoundedQueue`: Core implementation with backpressure and worker management
- `WorkQueue`: Higher-level wrapper with additional functionality
- `QueueConfig`: Configuration class for queue parameters
- Integration with the `Executor` class

### 1.2 Test Approach

We will use a combination of:

- **Unit Tests**: To verify the behavior of individual components in isolation
- **Integration Tests**: To verify the interaction between components
- **Mocking**: To isolate components from their dependencies
- **Parameterized Tests**: To test multiple scenarios efficiently

### 1.3 Key Testing Goals

- Verify that the queue correctly implements backpressure when full
- Verify that worker tasks are properly managed (creation, execution,
  cancellation)
- Verify that the queue properly handles errors in worker tasks
- Verify that resources are properly cleaned up when the queue is stopped
- Verify that the queue integrates correctly with the existing executor
  framework
- Achieve >80% code coverage for the new components

## 2. Test Environment

### 2.1 Test Framework

```
pytest
pytest-asyncio
pytest-mock
pytest-cov
```

### 2.2 Mock Framework

```
unittest.mock
pytest-mock
```

### 2.3 Test Database

Not applicable for this component.

## 3. Unit Tests

### 3.1 Test Suite: BoundedQueue Initialization

#### 3.1.1 Test Case: Initialization with Default Values

**Purpose:** Verify that BoundedQueue initializes with correct default values.

**Setup:**

```python
@pytest.fixture
def queue():
    return BoundedQueue(maxsize=10)
```

**Test Implementation:**

```python
def test_bounded_queue_initialization(queue):
    assert queue.maxsize == 10
    assert queue.timeout == 0.1
    assert queue.status == QueueStatus.IDLE
    assert queue.size == 0
    assert queue.is_empty
    assert not queue.is_full
    assert queue.metrics["enqueued"] == 0
    assert queue.metrics["processed"] == 0
    assert queue.metrics["errors"] == 0
    assert queue.metrics["backpressure_events"] == 0
```

#### 3.1.2 Test Case: Initialization with Custom Values

**Purpose:** Verify that BoundedQueue initializes with custom values.

**Test Implementation:**

```python
def test_bounded_queue_initialization_custom_values():
    queue = BoundedQueue(maxsize=5, timeout=0.5)
    assert queue.maxsize == 5
    assert queue.timeout == 0.5
```

#### 3.1.3 Test Case: Initialization with Invalid Maxsize

**Purpose:** Verify that BoundedQueue raises ValueError for invalid maxsize.

**Test Implementation:**

```python
def test_bounded_queue_initialization_invalid_maxsize():
    with pytest.raises(ValueError, match="Queue maxsize must be at least 1"):
        BoundedQueue(maxsize=0)
```

### 3.2 Test Suite: BoundedQueue Operations

#### 3.2.1 Test Case: Put and Get Operations

**Purpose:** Verify that put and get operations work correctly.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_bounded_queue_put_get():
    queue = BoundedQueue(maxsize=2)
    await queue.start()

    # Put items
    assert await queue.put("item1")
    assert await queue.put("item2")

    # Queue should be full now
    assert queue.is_full
    assert queue.size == 2

    # Get items
    item1 = await queue.get()
    queue.task_done()
    item2 = await queue.get()
    queue.task_done()

    assert item1 == "item1"
    assert item2 == "item2"
    assert queue.is_empty

    # Confirm metrics
    assert queue.metrics["enqueued"] == 2
    assert queue.metrics["processed"] == 2
    assert queue.metrics["errors"] == 0

    await queue.stop()
```

#### 3.2.2 Test Case: Backpressure

**Purpose:** Verify that the queue applies backpressure when full.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_bounded_queue_backpressure():
    queue = BoundedQueue(maxsize=1, timeout=0.01)
    await queue.start()

    # Put first item should succeed
    assert await queue.put("item1")

    # Second item should fail (backpressure)
    assert not await queue.put("item2")

    # Metrics should show backpressure event
    assert queue.metrics["backpressure_events"] == 1

    await queue.stop()
```

#### 3.2.3 Test Case: Join Operation

**Purpose:** Verify that join waits for all items to be processed.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_bounded_queue_join():
    queue = BoundedQueue(maxsize=10)
    await queue.start()

    # Put items
    await queue.put("item1")
    await queue.put("item2")

    # Process items in a separate task
    async def process_items():
        item1 = await queue.get()
        await asyncio.sleep(0.1)  # Simulate processing
        queue.task_done()

        item2 = await queue.get()
        await asyncio.sleep(0.1)  # Simulate processing
        queue.task_done()

    task = asyncio.create_task(process_items())

    # Join should wait for all items to be processed
    await queue.join()
    assert queue.is_empty

    # Cleanup
    await task
    await queue.stop()
```

### 3.3 Test Suite: BoundedQueue Lifecycle

#### 3.3.1 Test Case: Start and Stop

**Purpose:** Verify that start and stop methods change queue status correctly.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_bounded_queue_start_stop():
    queue = BoundedQueue(maxsize=10)

    # Initial state
    assert queue.status == QueueStatus.IDLE

    # Start
    await queue.start()
    assert queue.status == QueueStatus.PROCESSING

    # Stop
    await queue.stop()
    assert queue.status == QueueStatus.STOPPED
```

#### 3.3.2 Test Case: Operations on Non-Processing Queue

**Purpose:** Verify that operations raise errors when queue is not in PROCESSING
state.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_bounded_queue_operations_non_processing():
    queue = BoundedQueue(maxsize=10)

    # Operations should fail when queue is IDLE
    with pytest.raises(RuntimeError, match="Cannot put items when queue is idle"):
        await queue.put("item")

    with pytest.raises(RuntimeError, match="Cannot get items when queue is idle"):
        await queue.get()

    # Start and then stop the queue
    await queue.start()
    await queue.stop()

    # Operations should fail when queue is STOPPED
    with pytest.raises(RuntimeError, match="Cannot put items when queue is stopped"):
        await queue.put("item")

    with pytest.raises(RuntimeError, match="Cannot get items when queue is stopped"):
        await queue.get()
```

#### 3.3.3 Test Case: Async Context Manager

**Purpose:** Verify that the queue works correctly as an async context manager.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_bounded_queue_context_manager():
    async with BoundedQueue(maxsize=10) as queue:
        assert queue.status == QueueStatus.PROCESSING

        # Use the queue
        await queue.put("item")
        item = await queue.get()
        queue.task_done()

        assert item == "item"

    # Queue should be stopped after exiting context
    assert queue.status == QueueStatus.STOPPED
```

### 3.4 Test Suite: BoundedQueue Worker Management

#### 3.4.1 Test Case: Start Workers

**Purpose:** Verify that start_workers creates the specified number of workers.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_bounded_queue_start_workers():
    queue = BoundedQueue(maxsize=10)
    await queue.start()

    # Define a simple worker function
    async def worker(item):
        return item

    # Start workers
    await queue.start_workers(worker, num_workers=3)

    # Check worker count
    assert queue.worker_count == 3

    # Cleanup
    await queue.stop()
```

#### 3.4.2 Test Case: Workers Process Items

**Purpose:** Verify that workers process items from the queue.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_bounded_queue_workers_process_items():
    queue = BoundedQueue(maxsize=10)
    await queue.start()

    # Track processed items
    processed_items = []

    # Define a worker function
    async def worker(item):
        processed_items.append(item)

    # Start workers
    await queue.start_workers(worker, num_workers=2)

    # Add items to the queue
    for i in range(5):
        await queue.put(f"item{i}")

    # Wait for all items to be processed
    await queue.join()

    # Check that all items were processed
    assert len(processed_items) == 5
    assert set(processed_items) == {f"item{i}" for i in range(5)}

    # Cleanup
    await queue.stop()
```

#### 3.4.3 Test Case: Worker Error Handling

**Purpose:** Verify that workers handle errors gracefully.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_bounded_queue_worker_error_handling():
    queue = BoundedQueue(maxsize=10)
    await queue.start()

    # Define a worker function that raises an exception for certain items
    async def worker(item):
        if item == "error_item":
            raise ValueError("Test error")

    # Define an error handler
    error_items = []
    async def error_handler(error, item):
        error_items.append((error, item))

    # Start workers with error handler
    await queue.start_workers(worker, num_workers=1, error_handler=error_handler)

    # Add items to the queue, including one that will cause an error
    await queue.put("item1")
    await queue.put("error_item")
    await queue.put("item2")

    # Wait for all items to be processed
    await queue.join()

    # Check error handling
    assert len(error_items) == 1
    error, item = error_items[0]
    assert isinstance(error, ValueError)
    assert str(error) == "Test error"
    assert item == "error_item"

    # Check metrics
    assert queue.metrics["errors"] == 1
    assert queue.metrics["processed"] == 3  # All items should be marked as processed

    # Cleanup
    await queue.stop()
```

#### 3.4.4 Test Case: Stop Workers

**Purpose:** Verify that stop_workers cancels all worker tasks.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_bounded_queue_stop_workers():
    queue = BoundedQueue(maxsize=10)
    await queue.start()

    # Define a simple worker function
    async def worker(item):
        return item

    # Start workers
    await queue.start_workers(worker, num_workers=3)
    assert queue.worker_count == 3

    # Stop workers
    await queue.stop()
    assert queue.worker_count == 0
```

### 3.5 Test Suite: WorkQueue

#### 3.5.1 Test Case: WorkQueue Initialization

**Purpose:** Verify that WorkQueue initializes correctly.

**Test Implementation:**

```python
def test_work_queue_initialization():
    queue = WorkQueue(maxsize=10, timeout=0.5, concurrency_limit=5)
    assert queue.queue.maxsize == 10
    assert queue.queue.timeout == 0.5
    assert queue.concurrency_limit == 5
```

#### 3.5.2 Test Case: WorkQueue Delegation

**Purpose:** Verify that WorkQueue delegates operations to BoundedQueue.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_work_queue_delegation():
    # Create a mock BoundedQueue
    mock_queue = AsyncMock()
    mock_queue.put = AsyncMock(return_value=True)
    mock_queue.get = AsyncMock(return_value="item")
    mock_queue.join = AsyncMock()
    mock_queue.start = AsyncMock()
    mock_queue.stop = AsyncMock()
    mock_queue.start_workers = AsyncMock()

    # Create a WorkQueue with the mock
    queue = WorkQueue(maxsize=10)
    queue.queue = mock_queue

    # Test delegation
    await queue.start()
    mock_queue.start.assert_called_once()

    await queue.put("item")
    mock_queue.put.assert_called_once_with("item")

    await queue.join()
    mock_queue.join.assert_called_once()

    await queue.stop()
    mock_queue.stop.assert_called_once()

    await queue.process(worker_func=lambda x: x, num_workers=2)
    mock_queue.start_workers.assert_called_once()
```

#### 3.5.3 Test Case: WorkQueue Batch Processing

**Purpose:** Verify that batch_process handles a list of items correctly.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_work_queue_batch_process():
    queue = WorkQueue(maxsize=5, concurrency_limit=2)

    # Track processed items
    processed_items = []

    # Define a worker function
    async def worker(item):
        await asyncio.sleep(0.01)  # Simulate processing
        processed_items.append(item)

    # Process a batch of items
    items = [f"item{i}" for i in range(10)]
    await queue.batch_process(items, worker)

    # Check that all items were processed
    assert len(processed_items) == 10
    assert set(processed_items) == set(items)
```

#### 3.5.4 Test Case: WorkQueue Context Manager

**Purpose:** Verify that WorkQueue works correctly as an async context manager.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_work_queue_context_manager():
    # Create a mock BoundedQueue
    mock_queue = AsyncMock()
    mock_queue.start = AsyncMock()
    mock_queue.stop = AsyncMock()

    # Create a WorkQueue with the mock
    queue = WorkQueue(maxsize=10)
    queue.queue = mock_queue

    # Test context manager
    async with queue:
        mock_queue.start.assert_called_once()

    mock_queue.stop.assert_called_once()
```

### 3.6 Test Suite: QueueConfig

#### 3.6.1 Test Case: QueueConfig Validation

**Purpose:** Verify that QueueConfig validates parameters correctly.

**Test Implementation:**

```python
def test_queue_config_validation():
    # Valid configuration
    config = QueueConfig(queue_capacity=10, capacity_refresh_time=1.0, concurrency_limit=5)
    assert config.queue_capacity == 10
    assert config.capacity_refresh_time == 1.0
    assert config.concurrency_limit == 5

    # Invalid queue_capacity
    with pytest.raises(ValueError, match="Queue capacity must be at least 1"):
        QueueConfig(queue_capacity=0)

    # Invalid capacity_refresh_time
    with pytest.raises(ValueError, match="Capacity refresh time must be positive"):
        QueueConfig(capacity_refresh_time=0)

    # Invalid concurrency_limit
    with pytest.raises(ValueError, match="Concurrency limit must be at least 1"):
        QueueConfig(concurrency_limit=0)
```

## 4. Integration Tests

### 4.1 Test Suite: Integration with Executor

#### 4.1.1 Test Case: Executor with WorkQueue

**Purpose:** Verify that Executor correctly uses WorkQueue for task management.

**Setup:**

```python
@pytest.fixture
def mock_event():
    class MockEvent:
        def __init__(self, value):
            self.value = value
            self.execution = MagicMock()
            self.execution.status = "PENDING"

        async def invoke(self):
            await asyncio.sleep(0.01)  # Simulate processing
            self.execution.status = "COMPLETED"
            return self.value

    return MockEvent
```

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_executor_with_work_queue(mock_event):
    # Create queue config and executor
    queue_config = QueueConfig(queue_capacity=5, concurrency_limit=2)
    executor = Executor(event_type=mock_event, queue_config=queue_config)

    # Add events
    events = []
    async with executor:
        for i in range(10):
            event = mock_event(f"value{i}")
            events.append(event)
            executor.append(event)

        # Process all events
        await executor.forward()

        # Wait for completion
        while not executor.is_all_processed:
            await asyncio.sleep(0.1)

    # Verify all events were processed
    for event in events:
        assert event.execution.status == "COMPLETED"
```

#### 4.1.2 Test Case: Executor with Queue Backpressure

**Purpose:** Verify that Executor handles queue backpressure gracefully.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_executor_with_queue_backpressure(mock_event):
    # Create queue config with small capacity
    queue_config = QueueConfig(queue_capacity=2, concurrency_limit=1)
    executor = Executor(event_type=mock_event, queue_config=queue_config)

    # Create a slow event processor
    class SlowMockEvent(mock_event):
        async def invoke(self):
            await asyncio.sleep(0.1)  # Slow processing
            self.execution.status = "COMPLETED"
            return self.value

    # Add events
    events = []
    async with executor:
        for i in range(5):
            event = SlowMockEvent(f"value{i}")
            events.append(event)
            executor.append(event)

        # Process all events
        await executor.forward()

        # Wait for completion
        while not executor.is_all_processed:
            await asyncio.sleep(0.1)

    # Verify all events were processed despite backpressure
    for event in events:
        assert event.execution.status == "COMPLETED"
```

#### 4.1.3 Test Case: Executor Resource Cleanup

**Purpose:** Verify that Executor properly cleans up queue resources.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_executor_resource_cleanup(mock_event):
    # Create queue config and executor
    queue_config = QueueConfig(queue_capacity=5, concurrency_limit=2)
    executor = Executor(event_type=mock_event, queue_config=queue_config)

    # Mock the work_queue
    mock_work_queue = AsyncMock()
    mock_work_queue.start = AsyncMock()
    mock_work_queue.stop = AsyncMock()
    mock_work_queue.join = AsyncMock()
    mock_work_queue.put = AsyncMock(return_value=True)
    mock_work_queue.is_empty = True
    executor.work_queue = mock_work_queue

    # Use executor in context manager
    async with executor:
        # Simulate some work
        event = mock_event("test")
        executor.append(event)
        await executor.forward()

    # Verify resource cleanup
    mock_work_queue.join.assert_called()
    mock_work_queue.stop.assert_called_once()
```

## 5. Mock Implementation Details

### 5.1 Mock BoundedQueue

```python
class MockBoundedQueue:
    def __init__(self, maxsize=100, timeout=0.1):
        self.maxsize = maxsize
        self.timeout = timeout
        self.queue = []
        self.status = QueueStatus.IDLE
        self.metrics = {
            "enqueued": 0,
            "processed": 0,
            "errors": 0,
            "backpressure_events": 0,
        }
        self.worker_count = 0

    @property
    def size(self):
        return len(self.queue)

    @property
    def is_full(self):
        return self.size >= self.maxsize

    @property
    def is_empty(self):
        return self.size == 0

    async def start(self):
        self.status = QueueStatus.PROCESSING

    async def stop(self, timeout=None):
        self.status = QueueStatus.STOPPED
        self.worker_count = 0

    async def put(self, item):
        if self.status != QueueStatus.PROCESSING:
            raise RuntimeError(f"Cannot put items when queue is {self.status}")

        if self.is_full:
            self.metrics["backpressure_events"] += 1
            return False

        self.queue.append(item)
        self.metrics["enqueued"] += 1
        return True

    async def get(self):
        if self.status != QueueStatus.PROCESSING:
            raise RuntimeError(f"Cannot get items when queue is {self.status}")

        if self.is_empty:
            raise asyncio.QueueEmpty()

        return self.queue.pop(0)

    def task_done(self):
        self.metrics["processed"] += 1

    async def join(self):
        pass

    async def start_workers(self, worker_func, num_workers, error_handler=None):
        self.worker_count = num_workers

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()
```

## 6. Test Data

```python
# Test items
test_items = [
    "item1",
    "item2",
    "item3",
    "item4",
    "item5",
]

# Test events
test_events = [
    {"id": "1", "type": "test", "data": {"value": "test1"}},
    {"id": "2", "type": "test", "data": {"value": "test2"}},
    {"id": "3", "type": "test", "data": {"value": "test3"}},
]
```

## 7. Helper Functions

```python
async def wait_for_queue_empty(queue, timeout=1.0):
    """Wait for the queue to become empty with timeout."""
    start_time = time.monotonic()
    while not queue.is_empty:
        if time.monotonic() - start_time > timeout:
            raise TimeoutError("Timeout waiting for queue to become empty")
        await asyncio.sleep(0.01)
```

## 8. Test Coverage Targets

- **Line Coverage Target:** 90%
- **Branch Coverage Target:** 85%
- **Critical Modules:**
  - `BoundedQueue`: 95% coverage
  - `WorkQueue`: 90% coverage
  - `QueueConfig`: 100% coverage

## 9. Continuous Integration

```yaml
name: Test Queue Implementation
on: [push, pull_request]
jobs:
  tests:
    runs-on: ubuntu-latest
    steps:
      - uses: actions/checkout@v3
      - name: Set up Python
        uses: actions/setup-python@v4
        with:
          python-version: "3.10"
      - name: Install dependencies
        run: |
          python -m pip install --upgrade pip
          pip install -e ".[dev]"
      - name: Run tests
        run: |
          pytest tests/clients/test_queue.py --cov=src/khive/clients/queue --cov-report=xml
      - name: Upload coverage
        uses: codecov/codecov-action@v3
```

## 10. Notes and Caveats

### 10.1 Known Limitations

- The tests assume that asyncio.Queue behaves correctly, as we're building on
  top of it.
- Some edge cases around task cancellation might be difficult to test reliably.
- Race conditions in async code can be challenging to reproduce in tests.

### 10.2 Future Improvements

- Add more parameterized tests to cover a wider range of scenarios.
- Add stress tests to verify behavior under high load.
- Add tests for distributed queue scenarios (if needed in the future).
