---
title: Test Implementation Plan for Circuit Breaker and Retry Patterns
by: khive-implementer
created: 2025-05-18
updated: 2025-05-18
version: 1.0
doc_type: TI
output_subdir: ti
description: Comprehensive test plan for circuit breaker and retry patterns for resilient API operations
date: 2025-05-18
---

# Test Implementation Plan: Circuit Breaker and Retry Patterns

## 1. Overview

### 1.1 Component Under Test

This test plan covers the implementation of circuit breaker and retry patterns
for resilient API operations. The main components under test are:

1. Enhanced CircuitBreaker class in the resilience.py module
2. Improved retry_with_backoff function with additional configuration options
3. Integration with AsyncAPIClient for resilient HTTP requests
4. Integration with Endpoint class for resilient API calls

### 1.2 Test Approach

We will use a combination of:

- **Unit tests**: Testing individual components in isolation with mocked
  dependencies
- **Integration tests**: Testing the interaction between components with
  simulated failures
- **Performance tests**: Ensuring the resilience patterns don't introduce
  significant overhead

All tests will follow the TDD approach, with tests written before
implementation.

### 1.3 Key Testing Goals

1. Verify that circuit breaker correctly transitions between states (CLOSED,
   OPEN, HALF_OPEN)
2. Ensure retry mechanism correctly applies exponential backoff with jitter
3. Confirm proper resource cleanup during failures
4. Validate integration with existing components (AsyncAPIClient, Endpoint)
5. Achieve >80% test coverage for all new code

## 2. Test Environment

### 2.1 Test Framework

```
pytest
pytest-asyncio
pytest-mock
pytest-cov
pytest-httpserver (for HTTP server mocking)
```

### 2.2 Mock Framework

```
unittest.mock
pytest-mock
pytest-httpserver (for HTTP server mocking)
```

### 2.3 Test Database

No database is required for these tests. All external dependencies will be
mocked.

## 3. Unit Tests

### 3.1 Test Suite: CircuitBreaker

#### 3.1.1 Test Case: Initialization

**Purpose:** Verify that CircuitBreaker initializes with correct default values.

**Setup:**

```python
@pytest.fixture
def circuit_breaker():
    return CircuitBreaker(failure_threshold=5, recovery_time=30.0)
```

**Test Implementation:**

```python
def test_circuit_breaker_initialization(circuit_breaker):
    assert circuit_breaker.state == CircuitState.CLOSED
    assert circuit_breaker.failure_threshold == 5
    assert circuit_breaker.recovery_time == 30.0
    assert circuit_breaker.failure_count == 0
```

#### 3.1.2 Test Case: State Transition to OPEN

**Purpose:** Verify that CircuitBreaker transitions from CLOSED to OPEN after
reaching failure threshold.

**Setup:**

```python
@pytest.fixture
def failing_function():
    return AsyncMock(side_effect=ValueError("Test error"))
```

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_circuit_breaker_state_transition_to_open(failing_function):
    # Arrange
    cb = CircuitBreaker(failure_threshold=2)

    # Act & Assert
    # First failure - circuit stays closed
    with pytest.raises(ValueError):
        await cb.execute(failing_function)
    assert cb.state == CircuitState.CLOSED

    # Second failure - circuit opens
    with pytest.raises(ValueError):
        await cb.execute(failing_function)
    assert cb.state == CircuitState.OPEN

    # Call when circuit is open - raises CircuitBreakerOpenError
    with pytest.raises(CircuitBreakerOpenError):
        await cb.execute(failing_function)
```

#### 3.1.3 Test Case: State Transition to HALF_OPEN

**Purpose:** Verify that CircuitBreaker transitions from OPEN to HALF_OPEN after
recovery time.

**Setup:**

```python
@pytest.fixture
def mock_time():
    with patch('time.monotonic') as mock:
        yield mock
```

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_circuit_breaker_state_transition_to_half_open(failing_function, mock_time):
    # Arrange
    cb = CircuitBreaker(failure_threshold=1, recovery_time=60.0)
    mock_time.return_value = 100.0  # Initial time

    # Act & Assert
    # First failure - circuit opens
    with pytest.raises(ValueError):
        await cb.execute(failing_function)
    assert cb.state == CircuitState.OPEN

    # Time hasn't passed - circuit stays open
    with pytest.raises(CircuitBreakerOpenError):
        await cb.execute(failing_function)

    # Time passes - circuit transitions to half-open
    mock_time.return_value = 161.0  # 61 seconds later

    # Next call should be allowed (in half-open state)
    failing_function.side_effect = ValueError("Still failing")
    with pytest.raises(ValueError):
        await cb.execute(failing_function)
    assert cb.state == CircuitState.OPEN  # Failed in half-open, back to open
```

#### 3.1.4 Test Case: State Transition to CLOSED

**Purpose:** Verify that CircuitBreaker transitions from HALF_OPEN to CLOSED
after successful execution.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_circuit_breaker_state_transition_to_closed(mock_time):
    # Arrange
    cb = CircuitBreaker(failure_threshold=1, recovery_time=60.0)

    # Create a function that fails once then succeeds
    call_count = 0
    async def test_function():
        nonlocal call_count
        call_count += 1
        if call_count == 1:
            raise ValueError("First call fails")
        return "success"

    # Act & Assert
    # First call - circuit opens
    with pytest.raises(ValueError):
        await cb.execute(test_function)
    assert cb.state == CircuitState.OPEN

    # Time passes - circuit transitions to half-open
    mock_time.return_value = 100.0
    cb.last_failure_time = 0.0  # Force recovery time to pass

    # Next call succeeds - circuit closes
    result = await cb.execute(test_function)
    assert result == "success"
    assert cb.state == CircuitState.CLOSED
```

#### 3.1.5 Test Case: Excluded Exceptions

**Purpose:** Verify that excluded exceptions don't count toward failure
threshold.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_circuit_breaker_excluded_exceptions():
    # Arrange
    cb = CircuitBreaker(
        failure_threshold=2,
        excluded_exceptions=(ValueError,)
    )

    # Create a function that raises excluded exception
    async def test_function():
        raise ValueError("Excluded exception")

    # Act & Assert
    # Multiple excluded exceptions don't open circuit
    for _ in range(5):
        with pytest.raises(ValueError):
            await cb.execute(test_function)

    assert cb.state == CircuitState.CLOSED
```

#### 3.1.6 Test Case: Metrics Tracking

**Purpose:** Verify that CircuitBreaker correctly tracks metrics.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_circuit_breaker_metrics():
    # Arrange
    cb = CircuitBreaker(failure_threshold=2)

    # Create functions for success and failure
    async def success_function():
        return "success"

    async def failure_function():
        raise RuntimeError("Failure")

    # Act
    # Two successful calls
    await cb.execute(success_function)
    await cb.execute(success_function)

    # Two failed calls - opens circuit
    with pytest.raises(RuntimeError):
        await cb.execute(failure_function)
    with pytest.raises(RuntimeError):
        await cb.execute(failure_function)

    # Rejected call
    with pytest.raises(CircuitBreakerOpenError):
        await cb.execute(success_function)

    # Assert
    metrics = cb.metrics
    assert metrics["success_count"] == 2
    assert metrics["failure_count"] == 2
    assert metrics["rejected_count"] == 1
    assert len(metrics["state_changes"]) == 1
    assert metrics["state_changes"][0]["from"] == CircuitState.CLOSED
    assert metrics["state_changes"][0]["to"] == CircuitState.OPEN
```

### 3.2 Test Suite: Retry with Backoff

#### 3.2.1 Test Case: Successful Retry

**Purpose:** Verify that retry_with_backoff retries failed operations and
eventually succeeds.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_retry_with_backoff_success_after_failures():
    # Arrange
    call_count = 0

    async def test_function():
        nonlocal call_count
        call_count += 1
        if call_count < 3:
            raise ConnectionError("Test connection error")
        return "success"

    # Act
    result = await retry_with_backoff(
        test_function,
        retry_exceptions=(ConnectionError,),
        max_retries=3,
        base_delay=0.01
    )

    # Assert
    assert result == "success"
    assert call_count == 3
```

#### 3.2.2 Test Case: Max Retries Exceeded

**Purpose:** Verify that retry_with_backoff raises the last exception after max
retries.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_retry_with_backoff_max_retries_exceeded():
    # Arrange
    call_count = 0

    async def test_function():
        nonlocal call_count
        call_count += 1
        raise ConnectionError(f"Error on attempt {call_count}")

    # Act & Assert
    with pytest.raises(ConnectionError) as exc_info:
        await retry_with_backoff(
            test_function,
            retry_exceptions=(ConnectionError,),
            max_retries=3,
            base_delay=0.01
        )

    assert "Error on attempt 4" in str(exc_info.value)
    assert call_count == 4  # Initial attempt + 3 retries
```

#### 3.2.3 Test Case: Excluded Exceptions

**Purpose:** Verify that excluded exceptions are not retried.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_retry_with_backoff_excluded_exceptions():
    # Arrange
    call_count = 0

    async def test_function():
        nonlocal call_count
        call_count += 1
        if call_count == 1:
            raise ValueError("Excluded exception")
        return "success"

    # Act & Assert
    with pytest.raises(ValueError):
        await retry_with_backoff(
            test_function,
            retry_exceptions=(ConnectionError,),
            exclude_exceptions=(ValueError,),
            max_retries=3,
            base_delay=0.01
        )

    assert call_count == 1  # No retries for excluded exception
```

#### 3.2.4 Test Case: Backoff Timing

**Purpose:** Verify that retry_with_backoff applies correct exponential backoff.

**Setup:**

```python
@pytest.fixture
def mock_sleep():
    with patch('asyncio.sleep') as mock:
        mock.return_value = None
        yield mock
```

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_retry_with_backoff_timing(mock_sleep):
    # Arrange
    call_count = 0

    async def test_function():
        nonlocal call_count
        call_count += 1
        if call_count <= 3:
            raise ConnectionError(f"Error on attempt {call_count}")
        return "success"

    # Act
    await retry_with_backoff(
        test_function,
        retry_exceptions=(ConnectionError,),
        max_retries=3,
        base_delay=1.0,
        backoff_factor=2.0,
        jitter=False
    )

    # Assert
    assert call_count == 4
    assert mock_sleep.call_count == 3

    # Check sleep durations follow exponential pattern
    assert mock_sleep.call_args_list[0][0][0] == 1.0  # First retry: base_delay
    assert mock_sleep.call_args_list[1][0][0] == 2.0  # Second retry: base_delay * backoff_factor
    assert mock_sleep.call_args_list[2][0][0] == 4.0  # Third retry: base_delay * backoff_factor^2
```

#### 3.2.5 Test Case: Jitter

**Purpose:** Verify that retry_with_backoff applies jitter to backoff times.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_retry_with_backoff_jitter(mock_sleep):
    # Arrange
    call_count = 0

    async def test_function():
        nonlocal call_count
        call_count += 1
        if call_count <= 3:
            raise ConnectionError(f"Error on attempt {call_count}")
        return "success"

    # Act
    with patch('random.uniform', return_value=1.1) as mock_random:
        await retry_with_backoff(
            test_function,
            retry_exceptions=(ConnectionError,),
            max_retries=3,
            base_delay=1.0,
            backoff_factor=2.0,
            jitter=True
        )

    # Assert
    assert call_count == 4
    assert mock_sleep.call_count == 3

    # Check sleep durations include jitter
    assert mock_sleep.call_args_list[0][0][0] != 1.0
    assert mock_sleep.call_args_list[1][0][0] != 2.0
    assert mock_sleep.call_args_list[2][0][0] != 4.0
```

### 3.3 Test Suite: API Client Integration

#### 3.3.1 Test Case: API Client with Circuit Breaker

**Purpose:** Verify that circuit breaker can be applied to API client requests.

**Setup:**

```python
@pytest.fixture
def mock_client():
    client = AsyncMock()
    client.request.side_effect = [
        RuntimeError("First failure"),
        RuntimeError("Second failure"),
        AsyncMock()  # Success on third attempt
    ]
    return client
```

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_api_client_with_circuit_breaker(mock_client):
    # Arrange
    api_client = AsyncAPIClient(
        base_url="https://example.com",
        client=mock_client,
        circuit_breaker=CircuitBreaker(failure_threshold=2)
    )

    # Act & Assert
    # First failure
    with pytest.raises(RuntimeError):
        await api_client.request("GET", "/test")

    # Second failure - opens circuit
    with pytest.raises(RuntimeError):
        await api_client.request("GET", "/test")

    # Circuit is open - rejects request
    with pytest.raises(CircuitBreakerOpenError):
        await api_client.request("GET", "/test")
```

#### 3.3.2 Test Case: API Client with Retry

**Purpose:** Verify that retry can be applied to API client requests.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_api_client_with_retry(mock_client):
    # Arrange
    api_client = AsyncAPIClient(
        base_url="https://example.com",
        client=mock_client,
        retry_config=RetryConfig(
            max_retries=2,
            base_delay=0.01,
            retry_exceptions=(RuntimeError,)
        )
    )

    # Act
    await api_client.request("GET", "/test")

    # Assert
    assert mock_client.request.call_count == 3
```

### 3.4 Test Suite: Endpoint Integration

#### 3.4.1 Test Case: Endpoint with Circuit Breaker

**Purpose:** Verify that circuit breaker can be applied to Endpoint calls.

**Setup:**

```python
@pytest.fixture
def mock_endpoint_client():
    client = AsyncMock()
    return client

@pytest.fixture
def endpoint_config():
    return EndpointConfig(
        provider="test",
        endpoint="/test",
        transport_type="http"
    )
```

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_endpoint_with_circuit_breaker(mock_endpoint_client, endpoint_config):
    # Arrange
    mock_endpoint_client.request.side_effect = [
        RuntimeError("First failure"),
        RuntimeError("Second failure"),
        AsyncMock()  # Success on third attempt
    ]

    endpoint = Endpoint(
        config=endpoint_config,
        client=mock_endpoint_client,
        circuit_breaker=CircuitBreaker(failure_threshold=2)
    )

    # Act & Assert
    # First failure
    with pytest.raises(RuntimeError):
        await endpoint.call({"method": "GET"})

    # Second failure - opens circuit
    with pytest.raises(RuntimeError):
        await endpoint.call({"method": "GET"})

    # Circuit is open - rejects request
    with pytest.raises(CircuitBreakerOpenError):
        await endpoint.call({"method": "GET"})
```

## 4. Integration Tests

### 4.1 Test Suite: End-to-End Resilience

#### 4.1.1 Test Case: API Client with Failing HTTP Server

**Purpose:** Verify that API client handles intermittent failures with retry and
circuit breaker.

**Setup:**

```python
@pytest.fixture
def failing_http_server(httpserver):
    # Configure server to fail intermittently
    failure_count = 0

    def handler(request):
        nonlocal failure_count
        path = request.path

        if path == "/always-fail":
            return httpserver.respond_with_data("Server Error", status=500)

        if path == "/success-after-retry":
            failure_count += 1
            if failure_count <= 2:
                return httpserver.respond_with_data("Server Error", status=500)
            return httpserver.respond_with_json({"status": "success"})

        return httpserver.respond_with_json({"status": "success"})

    httpserver.expect_request("").respond_with_handler(handler)
    return httpserver
```

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_api_client_with_failing_server(failing_http_server):
    # Arrange
    base_url = failing_http_server.url_for("")
    client = AsyncAPIClient(
        base_url=base_url,
        timeout=0.5,
        circuit_breaker=CircuitBreaker(failure_threshold=3),
        retry_config=RetryConfig(
            max_retries=2,
            base_delay=0.01,
            retry_exceptions=(ServerError,)
        )
    )

    # Act & Assert
    # First call should retry but eventually succeed
    result = await client.get("/success-after-retry")
    assert result["status"] == "success"

    # Calls to consistently failing endpoint should eventually open circuit
    for _ in range(3):
        with pytest.raises(ServerError):
            await client.get("/always-fail")

    # Circuit should be open now
    with pytest.raises(CircuitBreakerOpenError):
        await client.get("/always-fail")
```

#### 4.1.2 Test Case: Resource Cleanup During Failures

**Purpose:** Verify that resources are properly cleaned up during failures.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_resource_cleanup_during_failures(failing_http_server):
    # Arrange
    base_url = failing_http_server.url_for("")
    client = AsyncAPIClient(
        base_url=base_url,
        timeout=0.5
    )

    # Track if client is closed
    original_close = client.close
    close_called = False

    async def tracked_close():
        nonlocal close_called
        close_called = True
        await original_close()

    client.close = tracked_close

    # Act
    try:
        async with client:
            await client.get("/always-fail")
    except ServerError:
        pass

    # Assert
    assert close_called, "Client close method was not called during exception"
```

## 5. Performance Tests

### 5.1 Benchmark / Load Testing

**Purpose:** Verify that resilience patterns don't introduce significant
overhead.

**Test Implementation:**

```python
def test_circuit_breaker_performance(benchmark):
    # Arrange
    cb = CircuitBreaker()

    async def success_function():
        return "success"

    # Act: Measure performance of circuit breaker execution
    async def run_with_circuit_breaker():
        return await cb.execute(success_function)

    # Assert: Circuit breaker should add minimal overhead
    result = benchmark(lambda: asyncio.run(run_with_circuit_breaker()))
    assert result == "success"
```

## 6. Error Handling Tests

### 6.1 Test Suite: Error Propagation

**Purpose:** Verify that errors are properly propagated and transformed.

**Test Implementation:**

```python
@pytest.mark.asyncio
async def test_error_propagation_in_circuit_breaker():
    # Arrange
    cb = CircuitBreaker()

    async def raise_error():
        raise ValueError("Original error")

    # Act & Assert
    with pytest.raises(ValueError) as exc_info:
        await cb.execute(raise_error)

    assert "Original error" in str(exc_info.value)
```

## 7. Mock Implementation Details

### 7.1 HTTP Server Mock

```python
class MockHTTPServer:
    def __init__(self, failure_pattern=None):
        """
        Initialize mock HTTP server.

        Args:
            failure_pattern: Optional pattern for simulating failures.
                None: No failures
                "intermittent": Fail every other request
                "rate_limit": Return 429 after 3 requests
                "timeout": Simulate timeouts
        """
        self.failure_pattern = failure_pattern
        self.request_count = 0
        self.requests = []

    async def handle_request(self, request):
        """Handle a request based on the failure pattern."""
        self.request_count += 1
        self.requests.append(request)

        if self.failure_pattern == "intermittent" and self.request_count % 2 == 0:
            return {"status": 500, "error": "Server Error"}

        if self.failure_pattern == "rate_limit" and self.request_count > 3:
            return {"status": 429, "error": "Rate Limit Exceeded", "retry-after": 30}

        if self.failure_pattern == "timeout" and self.request_count % 3 == 0:
            # Simulate timeout by delaying beyond client timeout
            await asyncio.sleep(10)

        return {"status": 200, "data": {"id": request.get("id", "unknown")}}
```

## 8. Test Data

```python
# Test requests
test_requests = [
    {"method": "GET", "url": "/api/test"},
    {"method": "POST", "url": "/api/test", "json": {"name": "Test Entity"}},
    {"method": "PUT", "url": "/api/test/1", "json": {"name": "Updated Entity"}},
    {"method": "DELETE", "url": "/api/test/1"}
]

# Test error responses
test_error_responses = [
    {"status": 400, "error": "Bad Request", "detail": "Invalid input"},
    {"status": 401, "error": "Unauthorized", "detail": "Authentication required"},
    {"status": 403, "error": "Forbidden", "detail": "Insufficient permissions"},
    {"status": 404, "error": "Not Found", "detail": "Resource not found"},
    {"status": 429, "error": "Too Many Requests", "detail": "Rate limit exceeded", "retry-after": 30},
    {"status": 500, "error": "Internal Server Error", "detail": "Server error"},
    {"status": 503, "error": "Service Unavailable", "detail": "Service temporarily unavailable"}
]
```

## 9. Helper Functions

```python
async def simulate_intermittent_failures(func, num_failures, *args, **kwargs):
    """
    Helper to simulate a function that fails a specific number of times then succeeds.

    Args:
        func: The async function to wrap
        num_failures: Number of times to fail before succeeding
        *args, **kwargs: Arguments to pass to func

    Returns:
        The result of func after num_failures failures
    """
    failure_count = 0

    async def wrapped():
        nonlocal failure_count
        if failure_count < num_failures:
            failure_count += 1
            raise ConnectionError(f"Simulated failure {failure_count}")
        return await func(*args, **kwargs)

    return wrapped
```

## 10. Test Coverage Targets

- **Line Coverage Target:** 90%
- **Branch Coverage Target:** 85%
- **Critical Modules:**
  - CircuitBreaker: 95% coverage
  - retry_with_backoff: 95% coverage
  - Integration with API client: 90% coverage
  - Integration with Endpoint: 90% coverage

## 11. Continuous Integration

```yaml
name: Test
on: [push, pull_request]
jobs:
  tests:
    runs-on: ubuntu-latest
    steps:
      - uses: actions/checkout@v3
      - name: Set up Python
        uses: actions/setup-python@v4
        with:
          python-version: "3.10"
      - name: Install dependencies
        run: |
          python -m pip install --upgrade pip
          pip install -r requirements-dev.txt
      - name: Run tests
        run: |
          pytest --cov=src tests/ --cov-report=xml
      - name: Upload coverage
        uses: codecov/codecov-action@v3
```

## 12. Notes and Caveats

### 12.1 Known Limitations

- The circuit breaker state is not persisted across application restarts
- The circuit breaker is not distributed, so each instance maintains its own
  state
- Retry with backoff may not be suitable for all types of operations, especially
  those with side effects

### 12.2 Future Improvements

- Add distributed circuit breaker state using Redis or similar
- Implement adaptive retry strategies based on response headers
- Add more sophisticated jitter algorithms for better load distribution
- Create a unified resilience configuration system for consistent behavior
  across components
