---
title: "Test Implementation: Refactor InfoService to use Endpoint instances"
issue: 97
author: "@khive-implementer"
created: "2025-05-18"
status: "Completed"
---

# Test Implementation: Refactor InfoService to use Endpoint instances

## 1. Overview

This document outlines the testing approach for the refactored `InfoService`
implementation that now uses `Endpoint` instances for all external API calls, as
specified in Issue #97 and TDS-80.md.

## 2. Test Objectives

- Verify that the refactored `InfoService` correctly uses `Endpoint` instances
- Ensure proper error handling for various failure scenarios
- Verify proper resource cleanup
- Achieve >80% test coverage for the refactored code

## 3. Test Strategy

### 3.1 Unit Tests

Unit tests will focus on testing individual components of the `InfoServiceGroup`
class:

1. **Initialization Tests**: Verify that the `InfoServiceGroup` initializes with
   `None` endpoints
2. **Search Method Tests**: Test the `_perplexity_search` and `_exa_search`
   methods with various inputs and mock endpoints
3. **Consult Method Tests**: Test the `_consult` method with various inputs and
   mock endpoints
4. **Error Handling Tests**: Test error handling for various failure scenarios
5. **Resource Cleanup Tests**: Test that resources are properly cleaned up in
   the `close()` method

### 3.2 Integration Tests

Integration tests will verify end-to-end functionality:

1. **Perplexity Search Integration**: Test end-to-end Perplexity search request
   handling
2. **Exa Search Integration**: Test end-to-end Exa search request handling
3. **Consult Integration**: Test end-to-end consult request handling

### 3.3 Mocking Strategy

To avoid making actual API calls during testing, we will use mocks:

1. Mock the `match_endpoint` function to return mock endpoints
2. Mock the provider-specific request classes
3. Mock the endpoint `call` method to return predefined responses or raise
   exceptions

## 4. Test Cases

### 4.1 Unit Tests

#### 4.1.1 Initialization Tests

- `test_info_service_initialization`: Verify that `InfoServiceGroup` initializes
  with `None` endpoints

#### 4.1.2 Perplexity Search Tests

- `test_perplexity_search_success`: Test successful Perplexity search
- `test_perplexity_search_error`: Test error handling for Perplexity search
- `test_perplexity_search_endpoint_none`: Test handling of `None` endpoint for
  Perplexity search

#### 4.1.3 Exa Search Tests

- `test_exa_search_success`: Test successful Exa search
- `test_exa_search_error`: Test error handling for Exa search
- `test_exa_search_endpoint_none`: Test handling of `None` endpoint for Exa
  search

#### 4.1.4 Model Call Tests

- `test_make_model_call_success`: Test successful model call
- `test_make_model_call_error`: Test error handling for model call

#### 4.1.5 Consult Tests

- `test_consult_success`: Test successful consult
- `test_consult_endpoint_none`: Test handling of `None` endpoint for consult

#### 4.1.6 Request Handling Tests

- `test_handle_request_perplexity_search`: Test handling of Perplexity search
  requests
- `test_handle_request_exa_search`: Test handling of Exa search requests
- `test_handle_request_consult`: Test handling of consult requests
- `test_handle_request_invalid_action`: Test handling of invalid actions

#### 4.1.7 Resource Cleanup Tests

- `test_close`: Test that resources are properly cleaned up in the `close()`
  method

### 4.2 Integration Tests

- `test_info_service_perplexity_search_integration`: Test end-to-end Perplexity
  search request handling
- `test_info_service_exa_search_integration`: Test end-to-end Exa search request
  handling
- `test_info_service_consult_integration`: Test end-to-end consult request
  handling

## 5. Test Implementation

The tests are implemented in the `tests/services/info/test_info_service.py`
file. The implementation follows these principles:

1. Use pytest fixtures for common setup
2. Use pytest-mock for mocking
3. Use pytest-asyncio for testing async code
4. Use clear, descriptive test names
5. Follow the Arrange-Act-Assert pattern

## 6. Test Results

All tests pass successfully, and the test coverage for the refactored code is

> 90%, exceeding the required 80%.

### 6.1 Test Coverage

```
Name                                             Stmts   Miss  Cover   Missing
------------------------------------------------------------------------------
src/khive/services/info/info_service.py             98     10    90%   36, 38, 92, 147, 174, 225-227, 242-243
```

## 7. Conclusion

The test implementation successfully verifies that the refactored `InfoService`
correctly uses `Endpoint` instances for all external API calls, handles errors
appropriately, and properly cleans up resources. The test coverage exceeds the
required 80%, providing confidence in the correctness of the implementation.
